from .kelerrors import *
from .kelenums import *


class ListStep(object):
    def __init__(self, current: float, current_slope: float, duration: float):
        self.current = float(current)
        self.current_slope = float(current_slope)
        self.duration = float(duration)


class LoadList(object):
    def __init__(self, save_slot: int, current_range: float, steps: [ListStep], loop_number: int):

        self.save_slot = int(save_slot)
        self.current_range = float(current_range)
        self.steps = steps
        self.loop_number = int(loop_number)

    def __str__(self):
        list_string = ":LIST {slot:d},{range:5.4f}A,{step_number:d},".format(slot=self.save_slot,
                                                                             range=self.current_range,
                                                                             step_number=len(self.steps))
        step_string = ""

        for s in self.steps:
            step_string = step_string + "{current:5.4f}A,{slope:5.4f}A/uS,{duration:5.4f}S,".format(current=s.current,
                                                                                                    slope=s.current_slope,
                                                                                                    duration=s.duration)

        list_string = list_string + "{steps}{loops}".format(steps=step_string, loops=self.loop_number)
        return list_string

    def validate(self):
        if 1 > self.save_slot or self.save_slot > 7:
            raise ValueError("save-slot can only be from 1-7")
        if len(self.steps) > 84:
            raise ValueError("a maximum of 84 steps is allowed")
        if len(self.steps) < 2:
            raise ValueError("a minimum of 2 steps is required")
        for s in self.steps:
            if s.current > self.current_range:
                raise ValueOutOfLimitError(s.current, self.current_range, "current set is out of current range")
            if s.current_slope > 1.5:
                raise ValueOutOfLimitError(s.current_slope, 1.5, "current slope is over device limit")


class OCPList(object):
    def __init__(self, save_slot: int, on_voltage: float, on_delay: float, current_range: float, initial_current: float,
                 step_current: float, step_delay: float, off_current: float, ocp_voltage: float, max_overcurrent: float,
                 min_overcurrent: float):

        self.save_slot = int(save_slot)
        self.on_voltage = float(on_voltage)
        self.on_delay = float(on_delay)
        self.current_range = float(current_range)
        self.initial_current = float(initial_current)
        self.step_current = float(step_current)
        self.step_delay = float(step_delay)
        self.off_current = float(off_current)
        self.ocp_voltage = float(ocp_voltage)
        self.max_overcurrent = float(max_overcurrent)
        self.min_overcurrent = float(min_overcurrent)

    def __str__(self):
        ocp_string = ":OCP {0:d},{1:5.4f}V,{2:5.4f}S,{3:5.4f}A,{4:5.4f}A,{5:5.4f}A,{6:5.4f}S,{7:5.4f}A,{8:5.4f}V,{9:5.4f}A,{10:5.4f}A".format(
            self.save_slot, self.on_voltage, self.on_delay, self.current_range, self.initial_current, self.step_current,
            self.step_delay, self.off_current, self.ocp_voltage, self.max_overcurrent, self.min_overcurrent)

        return ocp_string

    def validate(self):
        if 1 > self.save_slot or self.save_slot > 10:
            raise ValueError("save-slot can only be from 1-10")
        if self.initial_current > self.current_range:
            raise ValueOutOfLimitError(self.initial_current, self.current_range,
                                       "Current values must be below current range")
        if self.initial_current <= self.max_overcurrent:
            raise ValueError("initial current must be above max overcurrent")
        if self.off_current > self.min_overcurrent:
            raise ValueError("off current must be below other current values")
        if self.max_overcurrent <= self.min_overcurrent:
            raise ValueError("max overcurrent must be above min overcurrent")
        if self.step_current > self.initial_current:
            raise ValueError("current step must be same or lower than initial current")


class OPPList(object):
    def __init__(self, save_slot: int, on_voltage: float, on_delay: float, current_range: float, initial_power: float,
                 step_power: float, step_delay: float, off_power: float, opp_voltage: float, max_overpower: float,
                 min_overpower: float):

        self.save_slot = int(save_slot)
        self.on_voltage = float(on_voltage)
        self.on_delay = float(on_delay)
        self.current_range = float(current_range)
        self.initial_power = float(initial_power)
        self.step_power = float(step_power)
        self.step_delay = float(step_delay)
        self.off_power = float(off_power)
        self.opp_voltage = float(opp_voltage)
        self.max_overpower = float(max_overpower)
        self.min_overpower = float(min_overpower)

    def __str__(self):
        opp_string = ":OPP {0:d},{1:5.4f}V,{2:5.4f}S,{3:5.4f}A,{4:5.4f}W,{5:5.4f}W,{6:5.4f}S,{7:5.4f}W,{8:5.4f}V,{9:5.4f}W,{10:5.4f}W".format(
            self.save_slot, self.on_voltage, self.on_delay, self.current_range, self.initial_power, self.step_power,
            self.step_delay, self.off_power, self.opp_voltage, self.max_overpower, self.min_overpower)

        return opp_string

    def validate(self):
        if 1 > self.save_slot or self.save_slot > 10:
            raise ValueError("save-slot can only be from 1-10")
        if self.initial_power <= self.max_overpower:
            raise ValueError("initial power must be above max overpower")
        if self.off_power > self.min_overpower:
            raise ValueError("off power must be below other power values")
        if self.max_overpower <= self.min_overpower:
            raise ValueError("max overpower must be above min overpower")
        if self.step_power > self.initial_power:
            raise ValueError("power step must be same or lower than initial power")


class BattList(object):
    def __init__(self, save_slot: int, current_range: float, discharge_current: float, cutoff_voltage: float, cutoff_capacity: float, cutoff_time: float):

        self.save_slot = save_slot
        self.current_range = float(current_range)
        self.discharge_current = float(discharge_current)
        self.cutoff_voltage = float(cutoff_voltage)
        self.cutoff_capacity = float(cutoff_capacity)
        self.cutoff_time = float(cutoff_time)

    def __str__(self):
        batt_string = ":BATT {0:d},{1:5.4f}A,{2:5.4f}A,{3:5.4f}V,{4:5.4f}AH,{5:5.4f}M".format(self.save_slot, self.current_range, self.discharge_current, self.cutoff_voltage, self.cutoff_capacity, self.cutoff_time)

        return batt_string

    def validate(self):
        if 1 > self.save_slot or self.save_slot > 10:
            raise ValueError("save-slot can only be from 1-10")
        if self.discharge_current > self.current_range:
            raise ValueError("discharge current has to be lower or same as current range")


class CVList(object):
    def __init__(self, voltage1: float, voltage2: float, frequency: float, duty_cycle: float):
        self.function = Mode.dynamic_cv
        self.voltage1 = float(voltage1)
        self.voltage2 = float(voltage2)
        self.frequency = float(frequency)
        self.duty_cycle = float(duty_cycle)

    def __str__(self):
        cv_string = ":DYN 1,{0:5.4f}V,{1:5.4f}V,{2:5.4f}HZ,{3:5.4f}%".format(self.voltage1, self.voltage2, self.frequency, self.duty_cycle)

        return cv_string

    def validate(self, limit: float):
        if self.voltage2 > limit or self.voltage1 > limit:
            raise ValueOutOfLimitError(max(self.voltage2, self.voltage1), limit, "voltage value out of set limits")
        if self.duty_cycle >= 100:
            raise ValueOutOfLimitError(self.duty_cycle, 100, "duty cycle must be below 100%")


class CCList(object):
    def __init__(self, slope1: float, slope2: float, current1: float, current2: float, frequency: float, duty_cycle: float):
        self.function = Mode.dynamic_cc
        self.slope1 = float(slope1)
        self.slope2 = float(slope2)
        self.current1 = float(current1)
        self.current2 = float(current2)
        self.frequency = float(frequency)
        self.duty_cycle = float(duty_cycle)

    def __str__(self):
        cc_string = ":DYN 2,{0:5.4f}A/uS,{1:5.4f}A/uS,{2:5.4f}A,{3:5.4f}A,{4:5.4f}HZ,{5:5.4f}%".format(self.slope1, self.slope2, self.current1, self.current2, self.frequency, self.duty_cycle)

        return cc_string

    def validate(self, limit: float):
        if self.current2 > limit or self.current1 > limit:
            raise ValueOutOfLimitError(max(self.current2, self.current1), limit, "current value out of set limits")
        if self.duty_cycle >= 100:
            raise ValueOutOfLimitError(self.duty_cycle, 100, "duty cycle must be below 100%")
        if self.slope1 > 1.5:
            raise ValueOutOfLimitError(max(self.slope1, self.slope2), 1.5, "current slope is over device limit")


class CRList(object):
    def __init__(self, resistance1: float, resistance2: float, frequency: float, duty_cycle: float):
        self.function = Mode.dynamic_cr
        self.resistance1 = float(resistance1)
        self.resistance2 = float(resistance2)
        self.frequency = float(frequency)
        self.duty_cycle = float(duty_cycle)

    def __str__(self):
        cr_string = ":DYN 3,{0:5.4f}OHM,{1:5.4f}OHM,{2:5.4f}HZ,{3:5.4f}%".format(self.resistance1, self.resistance2, self.frequency, self.duty_cycle)

        return cr_string

    def validate(self, limit: float):
        if self.resistance2 > limit or self.resistance1 > limit:
            raise ValueOutOfLimitError(max(self.resistance2, self.resistance1), limit,
                                       "resistance value out of set limits")
        if self.duty_cycle >= 100:
            raise ValueOutOfLimitError(self.duty_cycle, 100, "duty cycle must be below 100%")


class CWList(object):
    def __init__(self, power1: float, power2: float, frequency: float, duty_cycle: float):
        self.function = Mode.dynamic_cw
        self.power1 = float(power1)
        self.power2 = float(power2)
        self.frequency = float(frequency)
        self.duty_cycle = float(duty_cycle)

    def __str__(self):
        cw_string = ":DYN 4,{0:5.4f}W,{1:5.4f}W,{2:5.4f}HZ,{3:5.4f}%".format(self.power1, self.power2, self.frequency, self.duty_cycle)

        return cw_string

    def validate(self, limit: float):
        if self.power2 > limit or self.power1 > limit:
            raise ValueOutOfLimitError(max(self.power2, self.power1), limit, "power value out of set limits")
        if self.duty_cycle >= 100:
            raise ValueOutOfLimitError(self.duty_cycle, 100, "duty cycle must be below 100%")


class PulseList(object):

    def __init__(self, slope1: float, slope2: float, current1: float, current2: float, duration: float):
        self.function = Mode.dynamic_pulse
        self.slope1 = float(slope1)
        self.slope2 = float(slope2)
        self.current1 = float(current1)
        self.current2 = float(current2)
        self.duration = float(duration)

    def __str__(self):
        pulse_string = ":DYN 5,{0:5.4f}A/uS,{1:5.4f}A/uS,{2:5.4f}A,{3:5.4f}A,{3:5.4f}S".format(self.slope1, self.slope2, self.current1, self.current2, self.duration)

        return pulse_string

    def validate(self, limit: float):
        if self.current2 > limit or self.current1 > limit:
            raise ValueOutOfLimitError(max(self.current2, self.current1), limit, "current value out of set limits")
        if self.slope1 > 1.5:
            raise ValueOutOfLimitError(max(self.slope1, self.slope2), 1.5, "current slope is over device limit")


class ToggleList(object):
    def __init__(self, slope1: float, slope2: float, current1: float, current2: float):
        self.function = Mode.dynamic_toggle
        self.slope1 = float(slope1)
        self.slope2 = float(slope2)
        self.current1 = float(current1)
        self.current2 = float(current2)

    def __str__(self):
        toggle_string = ":DYN 6,{0:5.4f}A/uS,{1:5.4f}A/uS,{2:5.4f}A,{3:5.4f}A".format(self.slope1, self.slope2, self.current1, self.current2)

        return toggle_string

    def validate(self, limit: float):
        if self.current2 > limit or self.current1 > limit:
            raise ValueOutOfLimitError(max(self.current2, self.current1), limit, "current value out of set limits")
        if self.slope1 > 1.5:
            raise ValueOutOfLimitError(max(self.slope1, self.slope2), 1.5, "current slope is over device limit")



"""
Serial communication with Korad KEL103 and potentially KEL102 electronic loads.

The intent is to give easy access to the electronic load,
eliminating the need to know special codes.

The object supports the python `with` statement to release the serial
port automatically:

from KELSerial import KELSerial

with KELSerial('/dev/tty.usbmodemfd121') as device:
    print("Model: ", device.model)
    print("Status: ", device.status)

LICENSE: MIT

RESOURCES:

This library has been initially copied from the py-korad-serial project(https://github.com/starforgelabs/py-korad-serial) and been modified to work with the electronic loads.


"""

from time import sleep
from .kellists import *
from .kelenums import *
from .kelerrors import *
import re
import serial
import ipaddress

# define Modes that support setting directly
settableModes = [Mode.constant_voltage, Mode.constant_current, Mode.constant_resistance, Mode.constant_power, Mode.short]


class Status(object):

    def __init__(self, status):
        """ Initialize object with a KELSerial status character.

        :param status: Status value
        :type status: int[]
        """
        super(Status, self).__init__()
        self.raw = status
        status_values = status.split(",")
        self.beep = OnOffState(int(status_values[0]))
        self.baudrate = BaudRate(int(status_values[1]))
        self.lock = OnOffState(int(status_values[2]))
        self.trigger = OnOffState(int(status_values[3]))
        self.comm = OnOffState(int(status_values[4]))
        # 1 more values is put out, unused according to documentation(though lock status and trigger status also is not mentioned and still available)

    def __repr__(self):
        return "{0}".format(self.raw)

    def __str__(self):
        message = "Beep: {0}, Lock: {1}, Baudrate: {2}, Trigger: {3}, Comm: {4}"
        return message.format(
            self.beep.name,
            self.lock.name,
            self.baudrate.b,
            self.trigger.name,
            self.comm.name
        )

    def __unicode__(self):
        return self.__str__()


def float_or_none(value):
    try:
        return float(value)
    except (TypeError, ValueError):
        return None


def on_off_setting_or_none(value):
    try:
        return OnOffState(value)
    except (TypeError, ValueError):
        return None


class KELSerial(object):
    """
    Wrapper for communicating with a KEL103(and possibly KEL102)
    electronic load as a serial interface.
    """

    class Memory(object):
        """ Wrap a memory setting. """

        def __init__(self, serial_, memory_number):
            super(KELSerial
                  .Memory, self).__init__()
            self.__serial = serial_
            self.number = memory_number

        def recall(self):
            """ Recall this memory's settings.  """
            self.__serial.send("*RCL {0}".format(self.number))

        def save(self):
            """ Save the current value to this memory. """
            self.__serial.send("*SAV {0}".format(self.number))

    class OnOffButton(object):
        """ Wrap an on/off button. """

        def __init__(self, serial_, on_command, off_command, get_command):
            super(KELSerial
                  .OnOffButton, self).__init__()
            self.__serial = serial_
            self._on = on_command
            self._off = off_command
            self._get = get_command

        def on(self):
            self.__serial.send(self._on)

        def off(self):
            self.__serial.send(self._off)

        def get(self):
            return on_off_setting_or_none(self.__serial.send_receive(self._get))

    class Serial(object):
        """ Serial operations.

        There are some quirky things in communication. They go here.
        """

        def __init__(self, port, rate=115200, debug=False, send_sleep_time=0.1):
            super(KELSerial
                  .Serial, self).__init__()

            self.send_sleep_time = send_sleep_time
            self.debug = debug
            self.port = serial.Serial(port, rate, timeout=1)

        def read_string(self, line_number=1):
            """ Read a string.

            It appears that the KEL103 returns
            newline terminated strings.

            :return: str
            """
            output = ""

            for line in range(1, line_number + 1):
                output += self.port.readline().decode()

            if self.debug:
                print("read: {0}".format(output))

            return output.strip('\n')

        def send(self, text):
            if self.debug:
                print("_send: ", text)

            text = "%s\n" % text
            self.port.write(text.encode('ascii'))

            sleep(self.send_sleep_time)  # may be needed, needs testing

        def send_receive(self, text, line_number=1):
            self.send(text)

            return self.read_string(line_number)

    def __init__(self, port, rate: BaudRate = 115200, debug=False, send_sleep_time=0.1):
        super(KELSerial, self).__init__()

        self.__serial = KELSerial.Serial(port, rate.b, debug, send_sleep_time)

        # Memory recall/save buttons 1 through 100 -> mapped to memories 0 to 99
        self.memories = [
            KELSerial
            .Memory(self.__serial, i) for i in range(1, 101)
        ]

        self.input = KELSerial.OnOffButton(self.__serial, ":INP ON", ":INP OFF", ":INP?")
        self.settings = KELSerial.Settings(self.__serial)

    class Settings(object):
        def __init__(self, serial_):
            super(KELSerial
                  .Settings, self).__init__()
            self.__serial = serial_

            self.beep = KELSerial.OnOffButton(self.__serial, ":SYST:BEEP ON", ":SYST:BEEP OFF", ":SYST:BEEP?")
            self.lock = KELSerial.OnOffButton(self.__serial, ":SYST:LOCK ON", ":SYST:LOCK OFF", ":SYST:LOCK?")
            self.dhcp = KELSerial.OnOffButton(self.__serial, ":SYST:DHCP 1", ":SYST:DHCP 0", ":SYST:DHCP?")
            self.trigger = KELSerial.OnOffButton(self.__serial, ":SYST:EXIT ON", ":SYST:EXIT OFF", ":SYST:EXIT?")
            self.compensation = KELSerial.OnOffButton(self.__serial, ":SYST:COMP ON", ":SYST:COMP OFF", ":SYST:COMP?")

            """
            Since changing limits seems to only work on upper limit there seems to be not much point of getting lower limit.
            This applies to all limits.
            """
        @property
        def current_limit(self):
            result = self.__serial.send_receive(":CURR:UPP?")
            return float_or_none(result.rstrip("A"))

        @current_limit.setter
        def current_limit(self, value):
            self.__serial.send(":CURR:UPP {0:5.4f}A".format(value))

        @property
        def voltage_limit(self):
            result = self.__serial.send_receive(":VOLT:UPP?")
            return float_or_none(result.rstrip("V"))

        @voltage_limit.setter
        def voltage_limit(self, value):
            self.__serial.send(":VOLT:UPP {0:5.4f}V".format(value))

        @property
        def resistance_limit(self):
            result = self.__serial.send_receive(":RES:UPP?")
            return float_or_none(result.rstrip("OHM"))

        @resistance_limit.setter
        def resistance_limit(self, value):
            self.__serial.send(":RES:UPP {0:5.4f}OHM".format(value))

        @property
        def power_limit(self):
            result = self.__serial.send_receive(":POW:UPP?")
            return float_or_none(result.rstrip("W"))

        @power_limit.setter
        def power_limit(self, value):
            self.__serial.send(":POW:UPP {0:5.4f}W".format(value))

        def factoryreset(self):
            self.__serial.send(":SYST:FACTRESET")

        @property
        def baudrate(self):
            result = self.__serial.send_receive(
                ":SYST:BAUD?"
            )
            return BaudRate(int(result))

        @baudrate.setter
        def baudrate(self, rate: BaudRate):
            self.__serial.send(":SYST:BAUD {0}".format(rate.b))

        @property
        def subnetmask(self):
            result = self.__serial.send_receive(
                ":SYST:SMASK?"
            )
            return str(result)

        @subnetmask.setter
        def subnetmask(self, value: str):
            ip = ipaddress.ip_address(value)
            self.__serial.send(":SYST:SMASK {0}".format(ip))

        @property
        def ipaddress(self):
            result = self.__serial.send_receive(
                ":SYST:IPAD?"
            )
            return str(result)

        @ipaddress.setter
        def ipaddress(self, value: str):
            ip = ipaddress.ip_address(value)
            self.__serial.send(":SYST:IPAD {0}".format(ip))

        @property
        def gateway(self):
            result = self.__serial.send_receive(
                ":SYST:GATE?"
            )
            return str(result)

        @gateway.setter
        def gateway(self, value: str):
            ip = ipaddress.ip_address(value)
            self.__serial.send(":SYST:GATE {0}".format(ip))

        @property
        def macaddress(self):
            result = self.__serial.send_receive(
                ":SYST:MAC?"
            )
            return str(result)

        @macaddress.setter
        def macaddress(self, value: str):
            """ verifying proper mac address """
            if re.match("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", value.lower()):
                self.__serial.send(":SYST:MAC {0}".format(value.replace(":", "-")))
            else:
                raise ValueError

        @property
        def port(self):
            result = self.__serial.send_receive(
                ":SYST:PORT?"
            )
            return int(result)

        @port.setter
        def port(self, value: int):
            self.__serial.send(":SYST:PORT {0}".format(value))

    def __enter__(self):
        """ See documentation for Python's ``with`` command.
        """
        return self

    def __exit__(self, _type, value, traceback):
        """ See documentation for Python's ``with`` command.
        """
        self.close()
        return False

    # ##################################################################
    # Serial operations
    # ##################################################################

    @property
    def is_open(self):
        """ Report whether the serial port is open.
        :rtype: bool
        """
        return self.__serial.port.isOpen()

    def close(self):
        """ Close the serial port """
        self.__serial.port.close()

    def open(self):
        """ Open the serial port """
        self.__serial.port.open()

    # ##################################################################
    # Load operations
    # ##################################################################

    def trigger(self):
        """Simulate an external trigger, used for Pulse and trigger dynamic mode"""
        self.__serial.send("*TRG")

    def set_list(self, load_list: LoadList, recall=True):
        load_list.validate()
        self.__serial.send(load_list.__str__())

        if recall:
            self.recall_list(load_list.save_slot)

    def recall_list(self, list_number: int):
        if 1 > list_number > 7:
            raise ValueError("save-slot can only be from 1-7")

        self.__serial.send(":RCL:LIST " + str(list_number))

    def get_list(self, list_number: int):

        if 1 > list_number > 7:
            raise ValueError("save-slot can only be from 1-7")

        self.recall_list(list_number)
        list_string = self.__serial.send_receive(":RCL:LIST?")
        if list_string == "":
            raise ValueError("No saved data in slot")
        list_string = list_string.replace(" ", "").replace("A/uS", "").replace("S", "").replace("A", "")

        split_string = list_string.split(",")
        current_range = float(split_string[0])
        loop_number = int(split_string[-1])
        split_string = split_string[2:-1]
        steps = []
        count = int(len(split_string) / 3)
        for s in range(count):
            steps.append(ListStep(float(split_string.pop(0)), float(split_string.pop(0)), float(split_string.pop(0))))

        return LoadList(list_number, current_range, steps, loop_number)

    def set_ocp(self, ocp_list: OCPList, recall=True):
        ocp_list.validate()

        self.__serial.send(ocp_list.__str__())

        if recall:
            self.recall_ocp(ocp_list.save_slot)

    def recall_ocp(self, list_number: int):
        if 1 > list_number > 10:
            raise ValueError("save-slot can only be from 1-10")

        self.__serial.send(":RCL:OCP " + str(list_number))

    def get_ocp(self, list_number: int):

        if 1 > list_number > 10:
            raise ValueError("save-slot can only be from 1-10")

        self.recall_ocp(list_number)
        list_string = self.__serial.send_receive(":RCL:OCP?")

        list_string = list_string.replace(" ", "").replace("A", "").replace("S", "").replace("V", "")

        split_string = list_string.split(",")
        on_voltage = float(split_string[0])
        on_delay = float(split_string[1])
        current_range = float(split_string[2])
        on_current = float(split_string[3])
        step_current = float(split_string[4])
        step_delay = float(split_string[5])
        off_current = float(split_string[6])
        ocp_voltage = float(split_string[7])
        max_overcurrent = float(split_string[8])
        min_overcurrent = float(split_string[9])

        return OCPList(list_number, on_voltage, on_delay, current_range, on_current, step_current, step_delay, off_current, ocp_voltage, max_overcurrent, min_overcurrent)

    def set_opp(self, opp_list: OPPList, recall=True):
        opp_list.validate()

        self.__serial.send(opp_list.__str__())

        if recall:
            self.recall_opp(opp_list.save_slot)

    def recall_opp(self, list_number: int):
        if 1 > list_number > 10:
            raise ValueError("save-slot can only be from 1-10")

        self.__serial.send(":RCL:OPP " + str(list_number))

    def get_opp(self, list_number: int):

        if 1 > list_number > 10:
            raise ValueError("save-slot can only be from 1-10")

        self.recall_opp(list_number)
        list_string = self.__serial.send_receive(":RCL:OPP?")

        list_string = list_string.replace(" ", "").replace("A", "").replace("S", "").replace("V", "").replace("W", "")

        split_string = list_string.split(",")
        on_voltage = float(split_string[0])
        on_delay = float(split_string[1])
        current_range = float(split_string[2])
        on_power = float(split_string[3])
        step_power = float(split_string[4])
        step_delay = float(split_string[5])
        off_power = float(split_string[6])
        opp_voltage = float(split_string[7])
        max_overpower = float(split_string[8])
        min_overpower = float(split_string[9])

        return OPPList(list_number, on_voltage, on_delay, current_range, on_power, step_power, step_delay, off_power,
                       opp_voltage, max_overpower, min_overpower)

    def set_batt(self, batt_list: BattList, recall=True):
        batt_list.validate()

        self.__serial.send(batt_list.__str__())

        if recall:
            self.recall_batt(batt_list.save_slot)

    def recall_batt(self, list_number: int):
        if 1 > list_number > 10:
            raise ValueError("save-slot can only be from 1-10")

        self.__serial.send(":RCL:BATT " + str(list_number))

    def get_batt(self, list_number: int):

        if 1 > list_number > 10:
            raise ValueError("save-slot can only be from 1-10")

        self.recall_batt(list_number)
        list_string = self.__serial.send_receive(":RCL:BATT?")

        list_string = list_string.replace(" ", "").replace("AH", "").replace("M", "").replace("V", "").replace("A", "")

        split_string = list_string.split(",")
        current_range = float(split_string[0])
        discharge_current = float(split_string[1])
        cutoff_voltage = float(split_string[2])
        cutoff_capacity = float(split_string[3])
        cutoff_time = float(split_string[4])

        return BattList(list_number, current_range, discharge_current, cutoff_voltage, cutoff_capacity, cutoff_time)

    def get_batt_time(self):
        batt_time = self.__serial.send_receive(":BATT:TIM?").replace("M", "")
        return float_or_none(batt_time)

    def get_batt_cap(self):
        batt_cap = self.__serial.send_receive(":BATT:CAP?").replace("AH", "")
        return float_or_none(batt_cap)

    def get_dynamic_mode(self):
        if self.function not in [Mode.dynamic_cv, Mode.dynamic_cc, Mode.dynamic_cr, Mode.dynamic_cw, Mode.dynamic_pulse,
                                 Mode.dynamic_toggle]:
            raise InvalidModeError(self.function)

        list_string = self.__serial.send_receive(":DYN?")
        list_string = list_string.replace(" ", "").replace("AH", "").replace("V", "").replace("A/uS", "").replace("HZ", "").replace("%", "").replace("A", "").replace("OHM", "").replace("W", "").replace("S", "")
        split_string = list_string.split(",")
        mode_string = int(split_string[0])

        match mode_string:
            case 1:
                voltage1 = float(split_string[1])
                voltage2 = float(split_string[2])
                frequency = float(split_string[3])
                duty_cycle = float(split_string[4])

                return CVList(voltage1, voltage2, frequency, duty_cycle)
            case 2:
                slope1 = float(split_string[1])
                slope2 = float(split_string[2])
                current1 = float(split_string[3])
                current2 = float(split_string[4])
                frequency = float(split_string[5])
                duty_cycle = float(split_string[6])

                return CCList(slope1, slope2, current1, current2, frequency, duty_cycle)
            case 3:
                resistance1 = float(split_string[1])
                resistance2 = float(split_string[2])
                frequency = float(split_string[3])
                duty_cycle = float(split_string[4])

                return CRList(resistance1, resistance2, frequency, duty_cycle)
            case 4:
                power1 = float(split_string[1])
                power2 = float(split_string[2])
                frequency = float(split_string[3])
                duty_cycle = float(split_string[4])

                return CWList(power1, power2, frequency, duty_cycle)
            case 5:
                slope1 = float(split_string[1])
                slope2 = float(split_string[2])
                current1 = float(split_string[3])
                current2 = float(split_string[4])
                duration = float(split_string[5])

                return PulseList(slope1, slope2, current1, current2, duration)
            case 6:
                slope1 = float(split_string[1])
                slope2 = float(split_string[2])
                current1 = float(split_string[3])
                current2 = float(split_string[4])

                return ToggleList(slope1, slope2, current1, current2)
            case _:
                raise InvalidModeError(list_string)

    def recall_dynamic_mode(self):
        return self.__serial.send_receive(":DYN?")

    def set_dynamic_mode(self, dynamic_list, recall=True):

        match dynamic_list.function:
            case Mode.dynamic_cv:
                limit = self.settings.voltage_limit
            case Mode.dynamic_cc:
                limit = self.settings.current_limit
            case Mode.dynamic_cr:
                limit = self.settings.resistance_limit
            case Mode.dynamic_cw:
                limit = self.settings.power_limit
            case Mode.dynamic_pulse:
                limit = self.settings.current_limit
            case Mode.dynamic_toggle:
                limit = self.settings.current_limit
            case _:
                raise ValueError()

        dynamic_list.validate(limit)

        self.__serial.send(dynamic_list.__str__())

        if recall:
            self.recall_dynamic_mode()

    @property
    def device_info(self):
        # Will just return full string, every item can already be gotten separately.
        """
        output is

        DHCP:0
        IP:192.168.1.198
        NETMASK:255.255.255.0
        GateWay:192.168.1.1
        MAC:70-2f-eb-48-4d-56
        PORT:18190
        BAUDRATE:115200
        """
        return self.__serial.send_receive(":SYST:DEVINFO?", 7)

    @property
    def model(self):
        """ Report the load model information.

        :rtype: str
        """
        return self.__serial.send_receive("*IDN?")

    @property
    def status(self):
        """ Report the load status.

        :rtype: KELSerial
    .Status or None
        """
        self.__serial.send(":STAT?")

        status = self.__serial.read_string()
        if len(status) == 0:
            return None
        else:
            return Status(status)

    @property
    def function(self):
        result = self.__serial.send_receive(
            ":FUNC?"
        )
        if len(result) == 0:
            return None
        else:
            return Mode(result)

    @function.setter
    def function(self, mode: Mode):
        if mode in settableModes:
            self.__serial.send(":FUNC {0}".format(mode.value))
        else:
            raise NoModeSetError(mode)

    @property
    def current(self):
        result = self.__serial.send_receive(
            ":CURR?"
        )
        return float_or_none(result.rstrip("A"))

    @current.setter
    def current(self, value):
        limit = self.settings.current_limit
        if value <= limit:
            self.__serial.send(":CURR {0:5.4f}A".format(value))
        else:
            raise ValueOutOfLimitError(value, limit)

    @property
    def voltage(self):
        return float_or_none(self.__serial.send_receive(":VOLT?").rstrip("V"))

    @voltage.setter
    def voltage(self, value):
        limit = self.settings.voltage_limit
        if value <= limit:
            self.__serial.send(":VOLT {0:5.4f}V".format(value))
        else:
            raise ValueOutOfLimitError(value, limit)

    @property
    def resistance(self):
        return float_or_none(self.__serial.send_receive(":RES?").rstrip("OHM"))

    @resistance.setter
    def resistance(self, value):
        limit = self.settings.resistance_limit
        if value <= limit:
            self.__serial.send(":RES {0:5.4f}OHM".format(value))
        else:
            raise ValueOutOfLimitError(value, limit)

    @property
    def power(self):
        return float_or_none(self.__serial.send_receive(":POW?").rstrip("W"))

    @power.setter
    def power(self, value):
        limit = self.settings.power_limit
        if value <= limit:
            self.__serial.send(":POW {0:5.4f}W".format(value))
        else:
            raise ValueOutOfLimitError(value, limit)

    @property
    def measured_current(self):
        """ Retrieve this device's currently sensed current.

        :return: Amperes
        :rtype: float or None
        """
        result = self.__serial.send_receive(
            ":MEAS:CURR?"
        )
        return float_or_none(result.rstrip("A"))

    @property
    def measured_voltage(self):
        """ Retrieve this device's currently sensed voltage.

        :return: Volts
        :rtype: float or None
        """
        result = self.__serial.send_receive(
            ":MEAS:VOLT?"
        )
        return float_or_none(result.rstrip("V"))

    @property
    def measured_power(self):
        """ Retrieve this device's used power.

        :return: Watts
        :rtype: float or None
        """
        result = self.__serial.send_receive(
            ":MEAS:POW?"
        )
        return float_or_none(result.rstrip("W"))



from aenum import Enum, MultiValueEnum


class Mode(Enum):
    """ Represents modes.

    These values should correspond to the values returned
    by the ``FUNCtion?`` command.
    """
    constant_voltage = "CV"
    constant_current = "CC"
    constant_resistance = "CR"
    constant_power = "CW"
    battery = "BATTERY"  # read only
    short = "SHORt"  # NOT a typo, actual output
    OCP = "OCP"  # read only
    LIST = "LIST"  # read only
    OPP = "OPP"  # read only
    dynamic_cv = "CONTINUOUS CV"  # read only - all dynamic modes
    dynamic_cc = "CONTINUOUS CC"
    dynamic_cr = "CONTINUOUS CR"
    dynamic_cw = "CONTINUOUS CW"
    dynamic_pulse = "PULSE"
    dynamic_toggle = "TOGGLE"


class BaudRate(MultiValueEnum):
    """
    These first values correspond to the baud rate returned
    by the ``STATUS?`` command. The scond values correspond to the baud rate returned by the ``:SYST:BAUD?`` command.
    """
    _init_ = 'a b'
    R9600 = (0, 9600)
    R19200 = (1, 19200)
    R38400 = (2, 38400)
    R57600 = (3, 57600)
    R115200 = (4, 115200)


class OnOffState(MultiValueEnum):
    """ Represents on/off states.

    This could just as easily be done as a Boolean, but is explicit.
    """
    _init_ = 'a b c'
    off = (0, "OFF", "0")
    on = (1, "ON", "1")



class InvalidModeError(Exception):
    """Exception raised when trying to get Mode that is not currently set
        Attributes:
            value -- string that was being tried to be converted
            message -- eplanation of the error
    """

    def __init__(self, mode, message="Got invalid mode from device"):
        self.mode = mode
        self.message = message
        super().__init__(self.message)


class ValueOutOfLimitError(Exception):
    """Exception raised for errors in the input value. Since there apparently exists a firmware bug allowing setting values above limits for resistance and power(see protocol docs) the library will check and limit values itself

    Attributes:
        value -- input value which caused the error
        limit -- the currently set limit
        message -- explanation of the error
    """

    def __init__(self, value, limit, message="Value is above limit"):
        self.value = value
        self.limit = limit
        self.message = message
        super().__init__(self.message)


class NoModeSetError(Exception):
    """Exception raised for trying to set mode that does not support it.

    Attributes:
        mode -- mode that was tried to be set
        message -- explanation of the error
    """

    def __init__(self, mode, message="Mode does not support setting directly from function"):
        self.mode = mode
        self.message = message
        super().__init__(self.message)
